package Lire::AsciiDlf::Records;

use strict;

use base qw/ Lire::AsciiDlf::ReportOperator Lire::Records /;

use Lire::Utils qw/tempfile/;
use Lire::DataTypes qw/ :basic :time format_numeric_type /;
use POSIX qw/ strftime /;
use Carp;

use vars qw/ $VERSION /;

BEGIN {
    ($VERSION)	= '$Revision: 1.15 $' =~ m!Revision: ([.\d]+)!;
}

#------------------------------------------------------------------------
# Method init_report( $period_start, $period_end )
#
# Method required by Lire::Ascii::ReportOperator
sub init_report {
    my $self = shift;

    # Cache index and formatters
    my @slices = ();
    foreach my $f ( @{$self->fields()} ) {
	my $field = $self->report_spec->schema->field( $f );
	push @slices, $field->pos;
    }
    $self->{slices}	= \@slices;

    $self->init_common;
    $self;
}

#------------------------------------------------------------------------
# Method init_merge($period_start, $period_end)
#
# Method required by Lire::AsciiDlf::ReportOperator
sub init_merge {
    my ( $self, $period_start, $period_end ) = @_;

    $self->init_common;

    $self;
}

#------------------------------------------------------------------------
# Method init_common
#
# Initialization common to both init_report() and init_merge()
sub init_common {
    my ( $self ) = @_;

    my @formatters = ();
    foreach my $f ( @{$self->fields()} ) {
	my $field = $self->report_spec->schema->field( $f );
	my $type = $field->type;
	if ( is_time_type( $type ) ) {
	    push @formatters, 
	      sub { strftime( '%Y-%m-%d %H:%M:%S', localtime $_[0] ) };
	} elsif ( is_quantity_type( $type ) ) {
	    push @formatters,
	      sub { format_numeric_type( $_[0], $type ) };
	} else {
	    push @formatters, sub { $_[0] };
	}
    }
    $self->{formatters} = \@formatters;

    # We use a temporary file to sort the DLF records according to
    # the group id we are using.
    $self->{fh}		= tempfile();
    $self->{group_no}	= 0;
}

#------------------------------------------------------------------------
# Method end_report()
#
# Method required by Lire::Ascii::ReportOperator
sub end_report {}

#------------------------------------------------------------------------
# Method init_group_data()
#
# Method required by Lire::AsciiDlf::ReportOperator
sub init_group_data {
    my ( $self ) = @_;

    my $value = $self->{group_no}++;

    return $value;
}

#------------------------------------------------------------------------
# Method update_group_data( $dlf, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub update_group_data {
    my ( $self, $dlf, $data ) = @_;

    my $fh = $self->{fh};
    print $fh join( " ", $data, map { defined $_ ? $_ : 'LIRE_NA' } @$dlf[@{$self->{slices}}] ), "\n";
}

#------------------------------------------------------------------------
# Method merge_group_data( $value, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub merge_group_data {
    my ( $self, $group, $data ) = @_;

    die "value should be of type Lire::Report::Group, not $group\n"
      unless UNIVERSAL::isa( $group, "Lire::Report::Group" );

    my $fh = $self->{fh};
    foreach my $e ( $group->entries ) {
	my @fields = $e->data;
	print $fh join( " ", $data, map { $_->{value}} @fields ), "\n";
    }
}

#------------------------------------------------------------------------
# Method end_group_data($data)
#
# Method required by Lire::AsciiDlf::ReportOperator
sub end_group_data {
    my ( $self, $data ) = @_;

    my $fh = $self->{fh};
    print $fh $data, " LIRE_RECORDS_GROUP_DATA_END\n";
}

#------------------------------------------------------------------------
# Method add_entry_value( $entry, $data)
#
# Method required by Lire::AsciiDlf::ReportOperator
sub add_entry_value {
    my ( $self, $entry, $data ) = @_;

    my $group = $entry->create_group;

    seek $self->{fh}, 0, 0
      or die "can't seek at the beginning of the temporary DLF file: $!\n";

    my $dlf_fh = $self->{fh};
    my $rx = qr/^$data /;
    while (<$dlf_fh>) {
	# Only keep the records matching the group id
	next unless /$rx/;

	chomp;
	my ( $id, @fields) = split / /, $_;

	# Check for end of group data marker
	last if $fields[0] eq 'LIRE_RECORDS_GROUP_DATA_END';

	my $entry = $group->create_entry;

	# FIXME: This needs to be revised once be get to 
	# the overhaul of the formatting engine. i.e. Multiple
	# names/values formatting.
	my $i = 0;
	foreach my $f ( @{$self->{formatters}} ) {
	    my $value = $fields[$i];
	    my $content = $f->( $value );
	    if ( $i == $#fields ) {
		$entry->add_value( content => $content,
                                   value => $value );
	    } else {
		$entry->add_name( $content, $value );
	    }
	    $i++;
	}
    }
}

# keep perl happy
1;

__END__

=pod

=head1 NAME

Lire::AsciiDlf::Records -

=head1 SYNOPSIS


=head1 DESCRIPTION

=head1 VERSION

$Id: Records.pm,v 1.15 2003/02/16 22:53:29 flacoste Exp $

=head1 COPYRIGHT

Copyright (C) 2001 Stichting LogReport Foundation LogReport@LogReport.org

This file is part of Lire.

Lire is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program (see COPYING); if not, check with
http://www.gnu.org/copyleft/gpl.html or write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.

=head1 AUTHOR

Francis J. Lacoste <flacoste@logreport.org>

=cut
