package Lire::FilterExpr;

use strict;

use Carp;

use vars qw/ $VERSION /;

BEGIN {
    ($VERSION)	= '$Revision: 1.14 $' =~ m!Revision: ([.\d]+)!;
}

sub new {
    my $proto = shift;
    my $class = ref $proto || $proto;

    my %args = @_;

    croak "container argument must be of type Lire::XMLSpecContainer"
      unless UNIVERSAL::isa( $args{container}, "Lire::XMLSpecContainer" );

    my $self = bless { container => $args{container},
		     }, $class;

    $self;
}

sub _is_value_a_ref {
    my ( $self, $value ) = @_;

    return substr( $value, 0, 1 ) eq '$';
}

sub _validate_ref_value {
    my ( $self, $value ) = @_;

    croak "value '$value' isn't a parameter or field reference"
      unless $self->_is_value_a_ref( $value );

    my $name = $self->_refname_from_value( $value );

    croak "$name isn't a defined field name or param name"
      unless $self->{container}->has_param( $name ) ||
        $self->{container}->schema()->has_field( $name );
}

sub _refname_from_value {
    my ( $self, $value ) = @_;

    croak "value '$value' isn't a parameter or field reference"
      unless $self->_is_value_a_ref( $value );

    return substr( $value, 1 );
}

sub _validate_value {
    my ( $self, $value ) = @_;

    $self->_validate_ref_value( $value )
      if $self->_is_value_a_ref( $value );
}

sub print {
    die ref($_[0])  . ": unimplemented abstract method";
}

package Lire::FilterExpr::BinaryExpr;

use Lire::Utils qw/xml_encode/;
use Carp;

use base qw/Lire::FilterExpr/;

sub new {
    my $self  = shift()->SUPER::new( @_ );

    my %args = @_;
    croak "missing 'arg1' argument to " . ref($self) . "::new"
      unless exists $args{arg1};
    croak "missing 'arg2' argument to " . ref($self) . "::new"
      unless exists $args{arg1};
    croak "missing 'op' argument to " . ref($self) . "::new"
      unless exists $args{arg1};

    $self->{op}	  = $args{op};
    $self->arg1(  $args{arg1} );
    $self->arg2(  $args{arg2} );

    $self;
}

sub arg1 {
    my ($self, $arg1) = @_;

    if ( @_ == 2) {
        croak "missing arg1 parameter"
          unless defined $arg1;
	$self->_validate_value( $arg1 );
	$self->{arg1} = $arg1;
    }

    $self->{arg1};
}

sub arg2 {
    my ($self, $arg2) = @_;

    if ( @_ == 2 ) {
        croak "missing arg2 parameter"
          unless defined $arg2;
	$self->_validate_value( $arg2 );
	$self->{arg2} = $arg2;
    }

    $self->{arg2};
}

sub print {
    my ($self,$fh, $prefix) = @_;
    $fh	    ||= \*STDOUT;
    $prefix ||= 0;

    my $pfx = " " x $prefix;
    my $arg1 = xml_encode( $self->{arg1} );
    my $arg2 = xml_encode( $self->{arg2} );

    print $fh $pfx,
      qq{<lire:$self->{op} arg1="$arg1" arg2="$arg2"/>\n};
}

package Lire::FilterExpr::Eq;

use base qw/ Lire::FilterExpr::BinaryExpr /;

sub new {
    return shift()->SUPER::new( @_ , op => "eq" );
}

package Lire::FilterExpr::Ne;

use base qw/ Lire::FilterExpr::BinaryExpr /;

sub new {
    return shift()->SUPER::new( @_ , op => "ne" );
}

package Lire::FilterExpr::BinaryNumericExpr;

use base qw/ Lire::FilterExpr::BinaryExpr /;

use Lire::DataTypes qw/ check_number is_numeric_type /;
use Carp;

sub _validate_value {
    my ( $self, $value ) = @_;

    if ( $self->_is_value_a_ref( $value )) {
        $self->_validate_ref_value( $value );

	my $name = $self->_refname_from_value( $value );
	my $type;
	# Make sure it has the proper type
	if ( $self->{container}->has_param( $name ) ) {
	    $type = $self->{container}->param( $name )->type;
	} else {
	    $type = $self->{container}->schema()->field( $name )->type;
	}
	croak "variable must be a numeric type"
	  unless is_numeric_type( $type );
    } else {
	croak "literal isn't an int or number"
	  unless check_number( $value );
    }
}

package Lire::FilterExpr::Lt;

use base qw/ Lire::FilterExpr::BinaryNumericExpr /;

sub new {
    return shift()->SUPER::new( @_ , op => "lt" );
}

package Lire::FilterExpr::Le;

use base qw/ Lire::FilterExpr::BinaryNumericExpr /;

sub new {
    return shift()->SUPER::new( @_ , op => "le" );
}

package Lire::FilterExpr::Gt;

use base qw/ Lire::FilterExpr::BinaryNumericExpr /;

sub new {
    return shift()->SUPER::new( @_ , op => "gt" );
}

package Lire::FilterExpr::Ge;

use base qw/ Lire::FilterExpr::BinaryNumericExpr /;

use Lire::Utils qw/xml_encode/;

sub new {
    return shift()->SUPER::new( @_ , op => "ge" );
}

package Lire::FilterExpr::Match;

use base qw/ Lire::FilterExpr /;

use Lire::DataTypes qw/ check_bool eval_bool /;
use Lire::Utils qw/xml_encode/;
use Carp;

sub new {
    my $self  = shift()->SUPER::new( @_ );

    my %args = @_;
    croak "missing 'value' argument to " . ref($self) . "::new"
      unless exists $args{value};
    croak "missing 're' argument to " . ref($self) . "::new"
      unless exists $args{re};

    $self->value( $args{value} );
    $self->re( $args{re} );
    $self->case_sensitive( $args{'case-sensitive'} || 0 );

    $self;
}

sub value {
    my ( $self, $value ) = @_;

    if ( @_ == 2) {
        croak "missing value parameter"
          unless defined $value;

	$self->_validate_value( $value );
	$self->{value} = $value;
    }

    $self->{value};
}

sub re {
    my ( $self, $re ) = @_;

    if ( @_ == 2 ) {
        croak "missing re parameter"
          unless defined $re;

        if ( $self->_is_value_a_ref( $re ) ) {
            my $name = $self->_refname_from_value( $re );
            $self->_validate_ref_value( $re );
            croak "re parameter cannot be a reference to a DLF field: $re"
              unless $self->{container}->has_param( $name );
        }
	$self->{re} = $re;
    }

    $self->{re};
}

sub case_sensitive {
    my ( $self, $cs ) = @_;

    if ( @_ == 2 ) {
        croak "missing cs parameter"
          unless defined $cs;

	croak "invalid bool value : $cs"
	  unless check_bool( $cs );
	$self->{case_sensitive} = $cs;
    }

    eval_bool( $self->{case_sensitive});
}

sub print {
    my ($self,$fh, $prefix) = @_;
    $fh	    ||= \*STDOUT;
    $prefix ||= 0;

    my $pfx = " " x $prefix;
    my $value = xml_encode( $self->{value} );
    my $re    = xml_encode( $self->{re} );

    print $fh $pfx, qq{<lire:match value="$value" re="$re"};
    if (defined $self->{case_sensitive}) {
	    print $fh qq{ case-sensitive="$self->{case_sensitive}"};
    }
    print $fh "/>\n";
}

package Lire::FilterExpr::Value;

use base qw/ Lire::FilterExpr /;

use Lire::Utils qw/xml_encode/;
use Carp;

sub new {
    my $self  = shift()->SUPER::new( @_ );

    my %args = @_;
    croak "missing 'value' argument to " . ref($self) . "::new"
      unless exists $args{value};

    $self->value( $args{value} );

    $self;
}

sub value {
    my ( $self, $value ) = @_;

    if ( @_ == 2 ) {
        croak "missing value parameter"
          unless defined $value;
	$self->_validate_value( $value );

	$self->{value} = $value;
    }

    $self->{value};
}

sub print {
    my ($self,$fh, $prefix) = @_;
    $fh	    ||= \*STDOUT;
    $prefix ||= 0;

    my $pfx = " " x $prefix;
    my $value = xml_encode( $self->{value} );
    print $fh $pfx, qq{<lire:value value="$value"/>\n};
}

package Lire::FilterExpr::Not;

use base qw/ Lire::FilterExpr /;

use Carp;

sub expr {
    my ( $self, $expr ) = @_;

    if ( @_ == 2 ) {
        croak "missing re parameter"
          unless defined $expr;
	croak "$expr isn't an expression"
	  unless UNIVERSAL::isa( $expr, 'Lire::FilterExpr' );
	$self->{expr} = $expr;
    }

    $self->{expr};
}

sub print {
    my ($self,$fh, $prefix) = @_;
    $fh	    ||= \*STDOUT;
    $prefix ||= 0;

    croak "invalid not expression: missing child"
      unless defined $self->{expr};

    my $pfx = " " x $prefix;
    print $fh $pfx, "<lire:not>\n";
    $self->{expr}->print( $fh, $prefix + 1 );
    print $fh $pfx, "</lire:not>\n";
}

package Lire::FilterExpr::And;

use base qw/ Lire::FilterExpr /;

use Carp;

sub expr {
    my ( $self, $expr ) = @_;

    if (@_ == 2 ) {
        croak "missing expr parameter"
          unless defined $expr;

	croak "$expr should be an array reference of expressions"
	  unless UNIVERSAL::isa( $expr, 'ARRAY' );

	croak "and must contains at leat one expression" unless @$expr;

	foreach my $e ( @$expr) {
	    croak "$e isn't an expression"
	      unless UNIVERSAL::isa( $e, 'Lire::FilterExpr' );
	}
	$self->{expr} = $expr;
    }

    $self->{expr};
}

sub print {
    my ($self,$fh, $prefix) = @_;
    $fh	    ||= \*STDOUT;
    $prefix ||= 0;

    croak "invalid and expression: missing children"
      unless ref $self->{expr} eq 'ARRAY' &&
	@{$self->{expr}};

    my $pfx = " " x $prefix;
    print $fh $pfx, "<lire:and>\n";
    foreach my $e ( @{$self->{expr}}) {
	$e->print( $fh, $prefix + 1 );
    }
    print $fh $pfx, "</lire:and>\n";
}

package Lire::FilterExpr::Or;

use base qw/ Lire::FilterExpr /;

use Carp;

sub expr {
    my ( $self, $expr ) = @_;

    if ( @_ == 2 ) {
        croak "missing 'expr' parameter"
          unless defined $expr;

	croak "$expr should be an array reference of expressions"
	  unless UNIVERSAL::isa( $expr, 'ARRAY' );
	croak "or must contains at leat one expression" unless @$expr;

	foreach my $e ( @$expr) {
	    croak "$e isn't an expression"
	      unless UNIVERSAL::isa( $e, 'Lire::FilterExpr' );
	}
	$self->{expr} = $expr;
    }

    $self->{expr};
}

sub print {
    my ($self,$fh, $prefix) = @_;
    $fh	    ||= \*STDOUT;
    $prefix ||= 0;

    croak "invalid or expression: missing children"
      unless ref $self->{expr} eq 'ARRAY' &&
	@{$self->{expr}};

    my $pfx = " " x $prefix;
    print $fh $pfx, "<lire:or>\n";
    foreach my $e ( @{$self->{expr}}) {
	$e->print( $fh, $prefix + 1 );
    }
    print $fh $pfx, "</lire:or>\n";
}

# keep perl happy
1;

__END__

=pod

=head1 NAME

Lire::FilterExpr -

=head1 SYNOPSIS


=head1 DESCRIPTION

=head1 VERSION

$Id: FilterExpr.pm,v 1.14 2003/02/03 00:31:47 flacoste Exp $

=head1 COPYRIGHT

Copyright (C) 2001 Stichting LogReport Foundation LogReport@LogReport.org

This file is part of Lire.

Lire is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program (see COPYING); if not, check with
http://www.gnu.org/copyleft/gpl.html or write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.

=head1 AUTHOR

Francis J. Lacoste <flacoste@logreport.org>

=cut
