package Lire::GroupOp;

use strict;

use base qw/ Lire::ReportOperator /;

use vars qw/ $VERSION /;

use Lire::ReportOperator;
use Carp;

BEGIN {
    ($VERSION)	= '$Revision: 1.21 $' =~ m!Revision: ([.\d]+)!;
}

=pod

=head1 NAME

Lire::GroupOp - Base class for all operators that compute values from a group
of DLF records

=head1 SYNOPSIS

    use Lire::GroupOp;

    use base qw/ Lire::GroupOp /;

=head1 DESCRIPTION

The Lire::GroupOp is the base class of all operators that computes
values from a group of DLF records. These are the operators that are
used inside aggregators.

This package also defines most of the classes that represents the
group operators available in Lire report specification.

=head1 METHODS FOR SUBCLASSES OF GroupOp

=head2 print( $fh, $prefix)

This methods implements the print() method required by
Lire::ReportOpetor. It prints an empty XML element named after op().
It also takes care of writing the name and label attribute. Other
attributes can be added to the XML element by overriding the
xml_attrs() method.

=cut

sub print {
    my ($self,$fh, $prefix) = @_;
    $fh	    ||= \*STDOUT;
    $prefix ||= 0;

    my $pfx = " " x $prefix;
    my $attrs = $self->xml_attrs;

    print $fh $pfx, "<lire:", $self->op, ' name="', $self->name, '"';
    print $fh qq{ label="$self->{label}"}
      if $self->{label};
    print $fh $attrs if length $attrs;
    print $fh "/>\n";
}

=pod

=head2 xml_attrs()

This method can be used to write additional XML attributes. The
returned string will be output in the XML element.

=cut

sub xml_attrs {
    return "";
}

#------------------------------------------------------------------------
# Method init()
#
# Make sure the parent parameter is set
sub init {
    my ( $self, %params ) = @_;

    croak "missing parent parameter\n"
      unless exists $params{parent};

    $self->SUPER::init( %params );
}

=pod

=head1 Lire::GroupOp::SimpleStat

This class provides common implementation for the operator that
implement common statistical operation on one DLF field.

Subclasses don't have to override xml_attrs() nor the print() method.


=cut

package Lire::GroupOp::SimpleStat;

use base qw/ Lire::GroupOp /;

use Lire::DataTypes qw/ :basic /;

use Carp;

=pod

=head2 init( %params )

Subclasses should provide an additional parameter to the init()
method.

=over

=item field

The DLF field's name on which the statistic is computed.

=back

=cut

sub init {
    my ( $self, %args ) = @_;

    $self->SUPER::init( %args );

    croak "missing 'field' argument"
      unless defined $args{field};

    $self->field(  $args{field} );

    $self;
}

=pod

=head2 field( [$new_field] )

Returns the name of the DLF field on which the statistic will be
computed.

If the $new_field is set, the field's attribute is changed to this new
value. It must be a valid field name for the schema of the current
report specification.

=cut

sub field {
    my ($self, $field) = @_;

    if ( @_ == 2 ) {
	croak "$field isn't a defined field in the ",
	  $self->report_spec->schema->id, " schema"
	    unless $self->report_spec->schema->has_field( $field );

	croak "$field isn't a numeric type",
	  unless is_numeric_type( $self->report_spec->schema->field( $field )->type );

	$self->{field} = $field;
    }

    $self->{field};
}

=pod

=head2 dlf_field()

Returns the field onto which we are computing a statistic as a
Lire::DlfField object.

=cut

sub dlf_field {
    $_[0]->report_spec->schema->field( $_[0]->field );
}

# ------------------------------------------------------------------------
# Method fill_group_info( $info )
#
# Implementation required by Lire::ReportOperator
sub fill_group_info {
    my ( $self, $info ) = @_;

    $info->create_column_info( $self->name, 'numerical',
			       $self->dlf_field->type, $self->label );
}

# ------------------------------------------------------------------------
# Method xml_attrs()
#
# Implementation required by Lire::GroupOp
sub xml_attrs {
    return qq{ field="$_[0]{field}"};
}

=pod

=head1 Lire::GroupOp::Count

Class that implements the count operator. This operator counts the
number of DLF records in the current group.

If the fields attribute is set, the count statistics will be equals to
the number of different values present in the fields specified by that
attribute.

Its also possible to express the count as a ratio of the total count
for the group or table.

=cut

package Lire::GroupOp::Count;

use base qw/ Lire::GroupOp /;

use Carp;

=pod

=head2 new( %params )

Creates a new Lire::GroupOp::Count object. In addition to the values
supported by its parents, the fields and ratio attributes will be
initialized to the values specified in the %params argument.

=cut

sub new {
    my $proto = shift;
    my $class = ref $proto || $proto;

    my $self  = bless { ratio => 'none', }, $class;

    $self->init( @_, op => "count" );

    my %args = @_;

    $self->fields(  $args{fields} )
      if exists $args{fields};

    $self->ratio( $args{ratio} )
      if exists $args{ratio};

    $self;
}

=pod

=head2 fields( [$new_fields] )

Returns the fields for which we are counting the different values.
This a reference to an array of DLF field names.

If the $new_fields parameter is set, it will be used as the new
$fields value. It must be an array reference and should only contains
valid field names for the current report specification's schema.

=cut

sub fields {
    my ($self, $fields) = @_;

    if ( @_ == 2 ) {
	if ( defined $fields ) {
	    croak "$fields isn't an array reference"
	      unless UNIVERSAL::isa( $fields, 'ARRAY' );

	    foreach my $f ( @$fields ) {
		croak "$f isn't a defined field in the ",
		  $self->report_spec->schema->id, " schema"
		    unless $self->report_spec->schema->has_field( $f );
	    }
	}
	$self->{fields} = $fields;
    }

    $self->{fields};
}

=pod

=head2 ratio([$new_ratio]) 

Returns how the count will be expressed. This can one of three
possible values:

=over

=item none

Default. The absolute count will be used.

=item group

The count will be expressed as a percentage of the group's count.

=item table

The count will be expressed as a percentage of the table's total
count.

=back

=cut

sub ratio {
    my ( $self, $ratio ) = @_;

    if ( @_ == 2 ) {
	croak "invalid value for ratio attribute: $ratio, should be one of none, group or table"
	  unless $ratio =~ /^(none|group|table)$/;

	$self->{ratio} = $ratio;
    }
    $self->{ratio};
}

#------------------------------------------------------------------------
# Method xml_attrs()
#
# Implementation needed by Lire::GroupOp
sub xml_attrs {
    my ($self) = @_;

    my $attr = "";
    if ( exists $self->{fields} ) {
	my $fields = join " ", @{$self->{fields}};
	$attr .= qq{fields="$fields"};
    }

    $attr .= qq{ ratio="$self->{ratio}"};

    $attr;
}

# ------------------------------------------------------------------------
# Method fill_group_info( $info )
#
# Implementation needed by Lire::ReportOperator
sub fill_group_info {
    my ( $self, $info ) = @_;

    $info->create_column_info( $self->name, 'numerical', 'int', $self->label );
}

=pod

=head1 Lire::GroupOp::Sum

Class that implements the sum operator. This operator will compute the
field's sum in a group of DLF records.

It's possible to compute a weighted sum in which each value is first
multiplied by the value of another DLF field.

Its also possible to express the count as a ratio of the total count
for the group or table.

=cut

package Lire::GroupOp::Sum;

use base qw/ Lire::GroupOp::SimpleStat /;

use Lire::DataTypes qw/ is_numeric_type /;

use Carp;

=pod

=head2 new( %params )

Creates a new Lire::GroupOp::Count object. In addition to the values
supported by its parents, the weight and ratio attributes will be
initialized to the values specified in the %params argument.

=cut

sub new {
    my $proto = shift;
    my $class = ref $proto || $proto;

    my $self  = bless { ratio => 'none', }, $class;

    $self->init( @_, op => "sum" );

    my %args = @_;
    $self->weight(  $args{weight} )
      if exists $args{weight};

    $self->ratio(  $args{ratio} )
      if exists $args{ratio};

    $self;
}

=pod

=head2 weigth( [$new_weight] )

Returns the DLF field's name by which the values will be multiplied
before being summed.

You can change the weight field by specifying a new name as the
$new_weight parameter. Use undef to remove the use of a weighting
field.

=cut

sub weight {
    my ($self, $weight) = @_;

    if ( @_ == 2 ) {
	if ( defined $weight ) {
	    croak "$weight isn't a defined field in the ",
	      $self->report_spec->schema->id, " schema"
		unless $self->report_spec->schema->has_field( $weight );

	    croak "$weight isn't a numeric type",
	      unless is_numeric_type( $self->report_spec->schema->field( $weight )->type );

	}

	$self->{weight} = $weight;
    }

    $self->{weight};
}

=pod

=head2 ratio([$new_ratio])

Returns how the sum will be expressed. This can one of three possible
values:

=over

=item none

Default. The absolute sum will be used.

=item group

The sum will be expressed as a percentage of the group's sum.

=item table

The sum will be expressed as a percentage of the table's total sum.

=back

=cut

sub ratio {
    my ( $self, $ratio ) = @_;

    if ( @_ == 2 ) {
	croak "invalid value for ratio attribute: $ratio, should be one of none, group or table"
	  unless $ratio =~ /^(none|group|table)$/;

	$self->{ratio} = $ratio;
    }
    $self->{ratio};
}

#------------------------------------------------------------------------
# Method xml_attrs()
#
# Implementation needed by Lire::GroupOp
sub xml_attrs {
    my ($self) = @_;

    my $attr = $self->SUPER::xml_attrs;

    $attr .= qq{ weight="$self->{weight}"}
      if defined $self->{weight};

    $attr .= qq{ ratio="$self->{ratio}"};

    $attr;
}

=pod

=head1 Lire::GroupOp::Avg

Class that implements the avg operator. This operator will compute the
average value of a DLF field in a group of DLF records.

If the field used by average isn't a numeric fields, the number of
different values in that field will be used as the numerator.

It's possible to compute a weighted average by specifying a weighting field.

By default, the average is a by record average. It's possible to
compute a by different values average. The list of fields that's used
to determine the different values if speficied in the by-fields
attribute.

=cut

package Lire::GroupOp::Avg;

use Carp;

use base qw/ Lire::GroupOp::SimpleStat /;

use Lire::DataTypes qw/ is_numeric_type /;

=pod

=head2 new( %params )

Creates a new Lire::GroupOp::Avg object. In addition to the values
supported by its parents, the by-fields and weight attributes will be
initialized to the values specified in the %params argument.

=cut

sub new {
    my $proto = shift;
    my $class = ref $proto || $proto;

    my $self  = bless {}, $class;

    $self->init( @_, op => "avg" );

    my %args = @_;
    $self->by_fields(  $args{'by-fields'} )
      if exists $args{'by-fields'};

    $self->weight(  $args{weight} )
      if exists $args{weight};

    $self;
}

#------------------------------------------------------------------------
# Method field( [$field]
#
# Overrides Lire::GroupOp::SimpleStat one since the field doesn't have
# to be numeric.
sub field {
    my ($self, $field ) = @_;

    if ( @_ == 2 ) {
	if ( defined $field ) {
	    croak "$field isn't a defined field in the ",
	      $self->report_spec->schema->id, " schema"
		unless $self->report_spec->schema->has_field( $field );
	}
	$self->{field} = $field;
    }

    $self->{field};
}

=pod

=head2 by_fields( [$new_by_fields] )

Returns the fields that are going to be used to count the different
values which will made up the by part of the average.

This a reference to an array of DLF field names.

If the $new_by_fields parameter is set, it will be used as the new
$by_fields value. It must be an array reference and should only
contains valid field names for the current report specification's
schema.

=cut

sub by_fields {
    my ($self, $by_fields) = @_;

    if ( @_ == 2 ) {
	if ( defined $by_fields ) {
	    croak "$by_fields isn't an array reference"
	      unless UNIVERSAL::isa( $by_fields, 'ARRAY' );

	    foreach my $f ( @$by_fields ) {
		croak "$f isn't a defined field in the ",
		  $self->report_spec->schema->id, " schema"
		    unless $self->report_spec->schema->has_field( $f );
	    }
	}
	$self->{by_fields} = $by_fields;
    }
    $self->{by_fields};
}

=pod

=head2 weigth( [$new_weight] )

Returns the DLF field's name by which the values will be multiplied
before being summed.

You can change the weight field by specifying a new name as the
$new_weight parameter. Use undef to remove the use of a weighting
field.

=cut

sub weight {
    my ($self, $weight) = @_;

    if ( @_ == 2 ) {
	if ( defined $weight ) {
	    croak "$weight isn't a defined field in the ",
	      $self->report_spec->schema->id, " schema"
		unless $self->report_spec->schema->has_field( $weight );

	    croak "$weight isn't a numeric type",
	      unless is_numeric_type( $self->report_spec->schema->field( $weight )->type );

	}

	$self->{weight} = $weight;
    }

    $self->{weight};
}

#------------------------------------------------------------------------
# Method xml_attrs()
#
# Implementation needed by Lire::GroupOp
sub xml_attrs {
    my ($self) = @_;

    my $attr = $self->SUPER::xml_attrs;
    $attr .= qq{ weight="$self->{weight}"}
      if defined $self->{weight};

    if ( exists $self->{by_fields} ) {
	my $by_fields = join " ", @{$self->{by_fields}};
	$attr .= qq{ by-fields="$by_fields"};
    }
    $attr;
}

=pod

=head1 Lire::GroupOp::Min

Class that implements the min operator. This operator will find the
minimum value appearing in a DLF field among a group of DLF records.

=cut

package Lire::GroupOp::Min;

use base qw/ Lire::GroupOp::SimpleStat /;

=pod

=head2 new( %params )

Creates a new Lire::GroupOp::Min object.

=cut

sub new {
    my $proto = shift;
    my $class = ref $proto || $proto;

    my $self  = bless {}, $class;

    $self->init( @_, op => "min" );
}

=pod

=head1 Lire::GroupOp::Max

Class that implements the max operator. This operator will find the
minimum value appearing in a DLF field among a group of DLF records.

=cut

package Lire::GroupOp::Max;

use base qw/ Lire::GroupOp::SimpleStat /;

=pod

=head2 new( %params )

Creates a new Lire::GroupOp::Max object.

=cut

sub new {
    my $proto = shift;
    my $class = ref $proto || $proto;

    my $self  = bless {}, $class;

    $self->init( @_, op => "max" );
}

=pod

=head1 Lire::GroupOp::First

Class that implements the first operator. This operator will output
the first value appearing in its field when the DLF records are sorted
according to the C<sort_fields> attribute. The default sort order is the
default timestamp sort.

=cut

package Lire::GroupOp::First;

use base qw/ Lire::GroupOp::SimpleStat /;

use Carp;

=pod

=head2 new( %params )

Creates a new Lire::GroupOp::First object. The field attribute can be
specified in the C<field> parameter and the sort order can be passed
in the C<sort_fields> parameter.

=cut

sub new {
    my $proto = shift;
    my $class = ref $proto || $proto;

    my $self  = bless {}, $class;

    my %params = @_;
    $self->init( @_, op => "first" );
    $self->sort_fields( $params{sort_fields} );
    $self;
}


#------------------------------------------------------------------------
# Method field( [$field]
#
# Overrides Lire::GroupOp::SimpleStat one since the field doesn't have
# to be numeric.
sub field {
    my ($self, $field ) = @_;

    if ( @_ == 2 ) {
	if ( defined $field ) {
	    croak "$field isn't a defined field in the ",
	      $self->report_spec->schema->id, " schema"
		unless $self->report_spec->schema->has_field( $field );
	}
	$self->{field} = $field;
    }

    $self->{field};
}

=pod

=head2 sort_fields( [$new_sort_fields] )

Returns the fields that are going to be used to sort the DLF records.
This a reference to an array of DLF field names.

If the $new_sort_fields parameter is set, it will be used as the new
sort order. It must be an array reference and should only contains
valid field names for the current report specification's schema.

When no sort_fields are set, the default is to use the default
timestamp field. =cut

=cut

sub sort_fields {
    my ($self, $sort_fields) = @_;

    if ( @_ == 2 ) {
	if ( defined $sort_fields ) {
	    croak "$sort_fields isn't an array reference"
	      unless UNIVERSAL::isa( $sort_fields, 'ARRAY' );

	    foreach my $f ( @$sort_fields ) {
		croak "$f isn't a defined field in the ",
		  $self->report_spec->schema->id, " schema"
		    unless $self->report_spec->schema->has_field( $f );
	    }
	}
	$self->{sort_fields} = $sort_fields;
    }
    $self->{sort_fields};
}

#------------------------------------------------------------------------
# Method xml_attrs()
#
# Implementation needed by Lire::GroupOp
sub xml_attrs {
    my ($self) = @_;

    my $attr = $self->SUPER::xml_attrs;
    if ( exists $self->{sort_fields} ) {
	my $sort_fields = join " ", @{$self->{sort_fields}};
	$attr .= qq{ sort="$sort_fields"};
    }
    $attr;
}

=pod

=head1 Lire::GroupOp::Last

Class that implements the first operator. This operator will output
the last value appearing in its field when the DLF records are sorted
according to the C<sort_fields> attribute. The default sort order is the
default timestamp sort.

=cut

package Lire::GroupOp::Last;

use base qw/ Lire::GroupOp::SimpleStat /;

use Carp;

=pod

=head2 new( %params )

Creates a new Lire::GroupOp::Last object.

=cut

sub new {
    my $proto = shift;
    my $class = ref $proto || $proto;

    my $self  = bless {}, $class;
    my %params = @_;

    $self->init( @_, op => "last" );
    $self->sort_fields( $params{sort_fields} );

    $self;
}

#------------------------------------------------------------------------
# Method field( [$field]
#
# Overrides Lire::GroupOp::SimpleStat one since the field doesn't have
# to be numeric.
sub field {
    my ($self, $field ) = @_;

    if ( @_ == 2 ) {
	if ( defined $field ) {
	    croak "$field isn't a defined field in the ",
	      $self->report_spec->schema->id, " schema"
		unless $self->report_spec->schema->has_field( $field );
	}
	$self->{field} = $field;
    }

    $self->{field};
}

=pod

=head2 sort_fields( [$new_sort_fields] )

Returns the fields that are going to be used to sort the DLF records.
This a reference to an array of DLF field names.

If the $new_sort_fields parameter is set, it will be used as the new
sort order. It must be an array reference and should only contains
valid field names for the current report specification's schema.

When no sort_fields are set, the default is to use the default
timestamp field. =cut

=cut

sub sort_fields {
    my ($self, $sort_fields) = @_;

    if ( @_ == 2 ) {
	if ( defined $sort_fields ) {
	    croak "$sort_fields isn't an array reference"
	      unless UNIVERSAL::isa( $sort_fields, 'ARRAY' );

	    foreach my $f ( @$sort_fields ) {
		croak "$f isn't a defined field in the ",
		  $self->report_spec->schema->id, " schema"
		    unless $self->report_spec->schema->has_field( $f );
	    }
	}
	$self->{sort_fields} = $sort_fields;
    }
    $self->{sort_fields};
}

#------------------------------------------------------------------------
# Method xml_attrs()
#
# Implementation needed by Lire::GroupOp
sub xml_attrs {
    my ($self) = @_;

    my $attr = $self->SUPER::xml_attrs;
    if ( exists $self->{sort_fields} ) {
	my $sort_fields = join " ", @{$self->{sort_fields}};
	$attr .= qq{ sort="$sort_fields"};
    }
    $attr;
}

# keep perl happy
1;

__END__

=head1 SEE ALSO

Lire::ReportSpec(3pm), Lire::ReportOperator(3pm), Lire::Aggregator(3pm).

=head1 VERSION

$Id: GroupOp.pm,v 1.21 2003/02/07 23:27:20 flacoste Exp $

=head1 COPYRIGHT

Copyright (C) 2001, 2002 Stichting LogReport Foundation LogReport@LogReport.org

This file is part of Lire.

Lire is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program (see COPYING); if not, check with
http://www.gnu.org/copyleft/gpl.html or write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.

=head1 AUTHOR

Francis J. Lacoste <flacoste@logreport.org>

=cut

