package Lire::Report::Group;

use strict;

use vars qw/ $VERSION @ISA /;

use Lire::Report::Entry;

use Carp;

BEGIN {
    ($VERSION)	= '$Revision: 1.14 $' =~ m!Revision: ([.\d]+)!;
}

=pod

=head1 NAME

Lire::Report::Group - Object that include subgroups and other entries.

=head1 SYNOPSIS

    foreach my $e ( $group->entries ) {
	# Process entries' data.
    }

=head1 DESCRIPTION

The Lire::Report::Group object contains entries grouped together in a
subreport.

=head1 CONSTRUCTOR

You create new Lire::Report::Group object by calling the create_group()
method on one Lire::Report::Entry object.

=cut

sub new {
    my $proto = shift;
    my $class = ref $proto || $proto;

    my ($entry, $group_info ) = @_;

    croak( "invalid parent_entry parameter: $entry" )
      unless UNIVERSAL::isa( $entry, 'Lire::Report::Entry' );

    croak( "invalid group_info parameter: $group_info" )
      unless UNIVERSAL::isa( $group_info, 'Lire::Report::GroupInfo' );

    bless {
	   type		 => 'group',
	   entries	 => [],
	   summary	 => {},
	   nrecords	 => 0,
           missing_cases => 0,
           show          => undef,
	   parent_entry	 => $entry,
	   group_info	 => $group_info,
	  }, $class;
}

=pod


=head1 OBJECT METHODS

=head2 parent_entry()

Returns the entry in which this goup is contained. This will be undef
for the Subreport since it's not contained in any entry.

=cut

sub parent_entry {
    $_[0]{parent_entry};
}

=pod

=head2 group_info()

Returns the Lire::Report::GroupInfo which contains the information
describing the entries of this group.

=cut

sub group_info {
    $_[0]{group_info}
}

=pod

=head2 subreport()

Returns the Lire::Report::Subreport object in which this entry is

=cut

sub subreport {
    $_[0]->parent_entry->subreport;
}

=pod

=head2 show( [$show] )

Returns the number of entries that should be displayed in the
formatted report. There may be more entries than that in the group.
When this parameter is zero or undefined, all available entries should
be displayed.

If the $show is parameter is set, the group's show parameter will
be set to this new value.

=cut

sub show {
    $_[0]{show} = $_[1] if @_ == 2;
    $_[0]{show};
}

=pod

=head2 nrecords( [$n] )

Returns the number of DLF records that were included in this group.

=cut

sub nrecords {
    if ( @_ == 2 ) {
	$_[0]{nrecords} = defined $_[1] ? $_[1] : 0;
    }

    $_[0]{nrecords};
}

=pod

=head2 missing_cases( [$n] )

Returns the number of DLF records that contained missing cases which
prevent them from being unclude in the subreport.

=cut

sub missing_cases {
    if ( @_ == 2 ) {
	$_[0]{missing_cases} = defined $_[1] ? $_[1] : 0;
    }

    $_[0]{missing_cases};
}

=pod

=head2 summary_values()

Returns the summary values of this group. This is an array of hash
reference like the ones returnes by the values() method described in
Lire::Report::Entry(3pm).

=cut

sub summary_values {
    my ( $self ) = @_;

    # Sort the columns according to column info's index.
    sort { $a->{col_info}->index <=> $b->{col_info}->index }
      values %{$self->{summary}};
}

=pod

=head2 get_summary_value( $name )

Returns the summary value of the operator $name. This is an hash
reference like is returned by the values() method described in
Lire::Report::Entry(3pm). Returns undef if the summary value wasn't
computed for operator $name.

=cut

sub get_summary_value {
    my ( $self, $name ) = @_;

    # Make sure that $name is defined in the the current table.

    return $self->{summary}{$name};
}

=pod

=head2 set_summary_value( $name, %value )

Sets the value of the operator $name computed over all the DLF records
of the group. Consult the add_value() method in
Lire::Report::Entry(3pm) for information on the parameters.

=cut

sub set_summary_value {
    my ( $self, $name, %value ) = @_;

    croak "missing name parameter"
      unless defined $name;

    $value{content}= ""
      unless defined $value{content};
    $value{value}= $value{content}
      unless defined $value{value};

    my $info = $self->{group_info}->column_info_by_name( $name );
    croak "there is no column named '$name'"
      unless $info;
    croak( "Operator $name is an aggregator" )
      unless $info->isa( "Lire::Report::ColumnInfo" );

    $info->update_chars_stats( $value{content} );
    $value{type} = 'value';
    $value{col_info} = $info;
    $value{n} = undef
      unless exists $value{n};
    $value{total} = undef
      unless exists $value{total};
    $value{missing_cases} = 0
      unless defined $value{missing_cases};

    $self->{summary}{$name} = \%value;
}

=pod

=head2 entries()

Returns the entries in the group.

=cut

sub entries {
    @{$_[0]{entries}};
}

=pod

=head2 create_entry()

Adds a data entry to this group. This will create a new
Lire::Report::Entry object which can then be filled with data.

=cut

sub create_entry {
    my ( $self ) = @_;

    croak( "can't create entries when group_info are missing" )
      unless $self->group_info;

    my $entry = new Lire::Report::Entry( $self );

    # Sets the row_idx to undef if this entry shouldn't be displayed
    if ( $self->show && $self->entries >= $self->show ) {
	$entry->row_idx( undef );
    } else {
	$entry->row_idx( $self->next_row_idx );
    }

    push @{$self->{entries}}, $entry;

    $entry;
}

# ------------------------------------------------------------------------
# Method next_row_idx()
#
# Returns the next row index which should be assigned to the next entry added 
# to this this group.
sub next_row_idx {
    my ( $self ) = @_;
    if ( @_ == 2 ) {
	$self->{next_row_idx} = $_[1];
	$self->{summary_row_idx} = $_[1];
	# The summary values sould appear above the first entry
	# in this group
	$self->{summary_row_idx} -= 1 if defined $_[1];
    }

    # The next_row_idx is only used for the first entry created
    my $last_entry;
    if ( $self->show && $self->entries >= $self->show ) {
	$last_entry = $self->{entries}[ $self->show - 1 ];
    } elsif ( $self->entries ) {
	$last_entry = $self->{entries}[ $#{$self->{entries}} ];
    } else {
	return $self->{next_row_idx};
    }

    $last_entry->next_row_idx;
}

# ------------------------------------------------------------------------
# Method write_report( [$fh, [$index]] )
#
sub write_report {
    my ( $self, $fh, $indent ) = @_;

    $fh ||= *STDOUT;
    my $pfx = ' ' x $indent;

    print $fh $pfx, "<lire:group";
    print $fh qq! show="$self->{show}"! 
      if defined $self->{show};
    print $fh ">\n";

    $self->write_group_summary( $fh, $indent + 1);

    foreach my $e ( $self->entries ) {
	$e->write_report( $fh, $indent + 1 );
    }

    print $fh $pfx, "</lire:group>\n";
}

# ------------------------------------------------------------------------
# write_group_summary( $fh, $indent )
#
# Write the group-summary element for this group
sub write_group_summary {
    my ( $self, $fh, $indent ) = @_;

    my $pfx = ' ' x $indent;
    print $fh $pfx, qq!<lire:group-summary nrecords="$self->{nrecords}" missing-cases="$self->{missing_cases}"!;
    print $fh qq{ row-idx="$self->{summary_row_idx}"}
      if defined $self->{summary_row_idx};
    print $fh ">\n";
    foreach my $v ( $self->summary_values ) {
	Lire::Report::Entry::write_value( $fh, $pfx . ' ', $v );
    }

    print $fh $pfx, "</lire:group-summary>\n";
}

#------------------------------------------------------------------------
# Method delete()
#
# Remove circular references
sub delete {
    my ( $self ) = @_;

    $self->{group_info}->delete;
    foreach my $e ( $self->entries ) {
	$e->delete;
    }
    %$self = ();
}

# keep perl happy
1;

__END__

=pod

=head1 SEE ALSO

Lire::ReportParser::ReportBuilder(3pm) Lire::Report(3pm)
Lire::Report::Subreport(3pm) Lire::Report::Entry(3pm)
Lire::Report::Image(3pm) Lire::Report::Section(3pm)

=head1 VERSION

$Id: Group.pm,v 1.14 2003/02/07 23:33:21 flacoste Exp $

=head1 COPYRIGHT

Copyright (C) 2002 Stichting LogReport Foundation LogReport@LogReport.org

This file is part of Lire.

Lire is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program (see COPYING); if not, check with
http://www.gnu.org/copyleft/gpl.html or write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.

=head1 AUTHOR

Francis J. Lacoste <flacoste@logreport.org>

=cut
