package Lire::ReportGenerator;

use strict;

use Lire::ReportConfig;
use Lire::Utils qw/ tempfile/ ;
use Lire::Logger;
use Lire::Config;
use Lire::DataTypes qw/ :basic :special :misc /;
use Lire::DlfStore;
use Lire::Utils qw/xml_encode tempdir/;
use Lire::AsciiDlf::AsciiDlfFactory;
use Lire::AsciiDlf::DlfInfo;

use POSIX qw/ strftime /;
use Symbol;
use Fcntl;
use Carp;

=pod

=head1 NAME

Lire::ReportGenerator - Generate a Report from a DLF file and a ReportConfig

=head1 SYNOPSIS

    use Lire::ReportGenerator;

    my $gen = new Lire::ReportGenerator( $superservice, $report_cfg_file,
                                         $dlf_file );
    my $report = $gen->generate_report;

=head1 DESCRIPTION

This method object will generate a report from a report configuration
file and a DLF file.

=cut

sub new {
    my $proto = shift;
    my $class = ref $proto || $proto;

    my ( $superservice, $report_cfg_file, $dlf_file ) = @_;

    croak "missing superservice parameter"
      unless defined $superservice;
    croak "missing report_cfg_file parameter"
      unless defined $report_cfg_file;
    croak "missing dlf_file parameter"
      unless defined $dlf_file;

    my $self = bless { superservice       => $superservice,
                       report_cfg_file    => $report_cfg_file,
                       dlf_file           => $dlf_file,
                       report_spec_by_key => {},
                       report_spec_order  => [],
                       dlf_store => undef,
                       start_time =>,
                       end_time =>,
                     }, $class;

    return $self;
}

sub import_dlf_in_store {
    my $self = $_[0];

    lr_info( "importing DLF into store..." );
    my $fh = gensym;
    open $fh, $self->{dlf_file}
      or die( "can't open $self->{dlf_file}: $!\n" );

    my $tmpdir = tempdir( "ReportGenerator_XXXXXX" );
    $self->{dlf_store} = Lire::DlfStore->open( "$tmpdir/dlf_store", 1  );
    my $schema = Lire::DlfSchema::load_schema( $self->{superservice} );
    my @fields = map { $_->name } $schema->fields;
    my $dlf_stream =
      $self->{dlf_store}->open_dlf_stream( $self->{superservice}, "w" );
    my $line;
    while (defined( $line = <$fh>)) {
        chomp;
        my @old_dlf = map { $_ eq 'LIRE_NOTAVAIL' ? undef : $_ }
          split /\s+/, $line;
        my $i = 0;
        my $dlf = { map { $_ => $old_dlf[$i++] } @fields };
        $dlf_stream->write_dlf( $dlf );
    }

    $self->{start_time} = $dlf_stream->start_time;
    $self->{end_time}   = $dlf_stream->end_time;
    $self->{nrecords}   = $dlf_stream->nrecords;

    $dlf_stream->close;
}

sub generate_report {
    my $self = $_[0];

    $self->load_report_cfg;
    $self->import_dlf_in_store;
    $self->print_dlf_stats;
    $self->create_other_dlf_streams;

    foreach my $stream ( $self->{dlf_store}->dlf_streams ) {
        $self->compute_reports_for_stream( $stream );
    }
    $self->{dlf_store}->close;

    lr_info( "creating lire XML report..." );
    return $self->{report_cfg}->create_report( $self->{start_time},
                                               $self->{end_time} );
}

sub load_report_cfg {
    my $self = $_[0];

    my $factory = new Lire::AsciiDlf::AsciiDlfFactory;
    $self->{report_cfg} =
      new_from_file Lire::ReportConfig( $self->{superservice},
                                        $self->{report_cfg_file},
                                        $factory);

    # Merge the sections' filter specifications
    $self->{report_cfg}->merge_filters;

    foreach my $section ( $self->{report_cfg}->sections ) {
	my @reports = $section->reports;
	foreach my $r ( @reports ) {
	    push @{$self->{report_spec_order}}, $r->key;
	    $self->{report_spec_by_key}{$r->key} = $r;
	}
    }
}

sub print_dlf_stats {
    my $self = $_[0];

    if ( ! defined $self->{start_time} ) {
	lr_info( "DLF contains ", $self->{nrecords},
                 " records; start and end time are unavailable" );
    } else {
	lr_info( "DLF contains ", $self->{nrecords},
                 " records; starts on ",
		 strftime( "%Y-%m-%d %H:%M:%S", localtime $self->{start_time}),
		 "; ends on ", 
		 strftime( "%Y-%m-%d %H:%M:%S", localtime $self->{end_time} ),
	       );
    }
}

sub create_other_dlf_streams {
    my $self = $_[0];

    foreach my $report ( @{$self->{report_spec_order}} ) {
	my $report_spec = $self->{report_spec_by_key}{$report};

        my $schema_id  = $report_spec->schema->id;
        unless ($self->{dlf_store}->has_dlf_stream( $schema_id ) ) {
            $self->create_dlf_stream( $schema_id );
        }
    }
}

sub create_dlf_stream {
    my ( $self, $schema_id ) = @_;

    my $schema = Lire::DlfSchema::load_schema( $schema_id );
    if ( $schema->can( 'base' ) ) {
        unless ($self->{dlf_store}->has_dlf_stream( $schema->base->id ) ) {
            $self->create_dlf_stream( $schema->base->id );
        }
    }
    eval {
	if ( UNIVERSAL::isa( $schema, "Lire::ExtendedSchema" ) ) {
	    $self->create_extended_dlf_stream( $schema );
	} elsif ( UNIVERSAL::isa( $schema, "Lire::DerivedSchema" ) ) {
	    $self->create_derived_dlf_stream( $schema );
	}
    };
    lr_warn( $@ )
      if $@;
}

sub init_reports_for_stream {
    my ($self, $stream_name) = @_;

    my $dlf_stream = $self->{dlf_store}->open_dlf_stream( $stream_name, "r" );
    my $period_start = $dlf_stream->start_time;
    my $period_end = $dlf_stream->end_time;
    $dlf_stream->close;
    foreach my $report ( @{$self->{report_spec_order}} ) {
	my $report_spec = $self->{report_spec_by_key}{$report};
	next unless defined $report_spec;
        next if $report_spec->schema->id ne $stream_name;

	lr_info( "initializing report '$report'");
	eval {
	    $report_spec->calc_spec->init_report( $period_start, $period_end );
	};
	if ( $@ ) {
	    lr_warn( "$@\n$report will be skipped");
            $self->cancel_one_report( $report, "failed: $@" );
	}
    }
}

sub reports_by_sortorders {
    my ($self, $stream_name) = @_;

    # An array used to group together report that share the same sort 
    # order on the input. This array contains hashes containing:
    #
    # - the fields on which to sort
    #    sort_fields => [], 
    # - the report's keys that don't have a filter expression
    #    reports	=> [],
    # - a hash used to group together report specs that share a common
    #   filter expression.
    #    filters    => {},
    my $sort_orders = [];
    foreach my $report ( @{$self->{report_spec_order}} ) {
	my $report_spec = $self->{report_spec_by_key}{$report};
	next unless defined $report_spec;
        next if $report_spec->schema->id ne $stream_name;

        $self->insert_report_in_sortorders( $report_spec, $sort_orders );
    }

    return $sort_orders;
}

sub insert_report_in_sortorders {
    my ( $self, $report_spec, $sort_orders ) = @_;

    my $report = $report_spec->key();

    # Put the report at the appropriate place in the data structure
    my @sfields = eval { $report_spec->calc_spec->dlf_sort_fields() };
    if ( $@ ) {
	lr_warn( "$@\n$report will be skipped");
        $self->cancel_one_report( $report, "dlf_sort_fields failed: $@" );
        return;
    }

    my $report_struct = undef;
    # Try to find an existing entry for those sort fields. An entry is 
    # considered compatible if they share a common sort fields prefix. 
    # For example, we can use the same sorted input for report 
    # specifications that need a sorted input in the following order:
    # time; time, client_host; time, client_host, http_result
    # No sort order is compatible with any sort order.
    foreach my $e ( @$sort_orders ) {
	my $order = $self->is_sort_order_compatible($e->{sort_fields}, \@sfields);
	next unless defined $order;

	$e->{sort_fields} = $order;
	$report_struct = $e;
	last;
    }
    unless ( $report_struct ) {
	$report_struct ||= {
			    sort_fields => \@sfields,
			    reports     => [],
			    filters     => {},
			   };
	push @$sort_orders, $report_struct;
    }

    if ( $report_spec->filter_spec ) {
	$self->insert_report_using_filter( $report_struct, $report_spec );
    } else {
	push @{$report_struct->{reports}}, $report;
    }
}

sub insert_report_using_filter {
    my ( $self, $rstruct, $report_spec ) = @_;

    my $filter_id = $report_spec->filter_spec->id;
    unless ( $rstruct->{filters}{$filter_id} ) {
        $rstruct->{filters}{$filter_id} = { name    => $filter_id,
                                            filter  => $report_spec->filter_spec->compile(),
                                            reports => [],
                                          };
    }
    push @{$rstruct->{filters}{$filter_id}{reports}}, $report_spec->key;
}

# Return the longest array when an array is a prefix of the other.
# Return undef when the arrays aren't equal or a prefix of one another.
sub is_sort_order_compatible {
    my ( $self, $sfield1, $sfield2 ) = @_;

    # Sfield1 should always contain the shortest array
    ($sfield1, $sfield2) = ($sfield2, $sfield1) if @$sfield2 < @$sfield1;
    for ( my $i=0; $i < @$sfield1; $i++ ) {
	return undef unless $sfield1->[$i] eq $sfield2->[$i];
    }

    # sfield1 is a prefix of sfield2
    return $sfield2;
}

sub cancel_reports {
    my ( $self, $report_struct ) = @_;

    foreach my $r ( @{$report_struct->{reports}} ) {
	lr_notice( "report '$r' will be skipped because of unavailable input");
        $self->cancel_one_report( $r, "unavailable input" );
    }

    foreach my $f ( values %{$report_struct->{filters}} ) {
	foreach my $r ( @{$f->{reports}} ) {
	    lr_notice( "report '$r' will be skipped because of unavailable input");
            $self->cancel_one_report( $r, "unavailable input" );
	}
    }
}

sub cancel_one_report {
    my ( $self, $report_key, $reason ) = @_;

    return unless $self->{report_spec_by_key}{$report_key};

    $self->{report_spec_by_key}{$report_key}->mark_missing( $reason );
    delete $self->{report_spec_by_key}{$report_key};
}

sub create_writer_cb {
    my ( $self, $schema, $out_stream ) = @_;

    my @fields = map { $_->name } $schema->fields;
    return sub {
        my $i = 0;
        my %dlf = map { $fields[$i++] => 
                          defined $_ && $_ ne 'LIRE_NOTAVAIL' ? $_ : undef }
          @{$_[0]};
        $out_stream->write_dlf( \%dlf );
    };
}

sub create_derived_dlf_stream {
    my ( $self, $schema ) = @_;
    lr_info( "creating derived DLF stream '", $schema->id, "' ",
	     "from '", $schema->base->id, "'" );

    my $dlf_stream =
      $self->{dlf_store}->open_dlf_stream( $schema->base->id, "r" );
    my $out_stream =
      $self->{dlf_store}->open_dlf_stream( $schema->id, "w" );

    my $writer_cb = $self->create_writer_cb( $schema, $out_stream );
    my $module = $schema->module;
    $module->init_computation( new Lire::AsciiDlf::DlfInfo( $dlf_stream ),
                               $writer_cb );

    my $dlf;
    my $count = 0;
    my $rtotal = $dlf_stream->nrecords;
    my $required_fields_checker = 
      $self->create_required_fields_checker_func( $schema );
    while ( $dlf = $dlf_stream->read_dlf_aref ) {
        next unless $required_fields_checker->( $dlf );
	$module->dlf_record( $dlf, $writer_cb );
	$count++;

	lr_info( sprintf( "%.2f%%", $count*100 / $rtotal ),
		 " of DLF source '", $schema->id, "' completed" )
	  unless $count % 10_000;
    }
    $module->end_computation( $writer_cb );
    lr_info( $out_stream->nrecords,
	     " records in '", $schema->id, "' derived schema" );
    $out_stream->close;
    $dlf_stream->close;
}

sub create_extended_dlf_stream {
    my ( $self, $schema ) = @_;

    lr_info( "creating extended DLF source '", $schema->id ,"' ",
		 "from '", $schema->base->id, "'" );

    my $dlf_stream =
      $self->{dlf_store}->open_dlf_stream( $schema->base->id, "r" );
    my $out_stream =
      $self->{dlf_store}->open_dlf_stream( $schema->id, "w" );

    my $module = $schema->module;
    $module->init_computation( new Lire::AsciiDlf::DlfInfo( $dlf_stream ) );

    my $dlf;
    my $count = 0;
    my $rtotal	= $dlf_stream->nrecords;
    my $required_fields_checker = 
      $self->create_required_fields_checker_func( $schema );
    my @fields = map { $_->name } $schema->fields;
    while ($dlf = $dlf_stream->read_dlf_aref ) {
        next unless $required_fields_checker->( $dlf );
	my $fields = $module->create_extended_fields( $dlf );

        my $i = 0;
        my %r = map { $fields[$i++] => 
                        defined $_ && $_ ne 'LIRE_NOTAVAIL' ? $_ : undef }
          @$dlf, @$fields;
        $out_stream->write_dlf( \%r );
	$count++;
	lr_info( sprintf( "%.2f%%", $count*100 / $rtotal ),
		 " of DLF source '", $schema->id, "' completed" )
	  unless $count % 10_000;
    }
    $module->end_computation;
    lr_info( $out_stream->nrecords, " records in '", $schema->id, 
             "' extended schema" );

    $out_stream->close;
    $dlf_stream->close;
}

sub create_required_fields_checker_func {
    my ( $self , $schema ) = @_;

    my @fields_pos = ();
    foreach my $field ( @{$schema->required_fields} ) {
        push @fields_pos, $schema->base->field( $field )->pos;
    }
    return sub {
        foreach my $f ( @fields_pos ) {
            return 0 unless defined $_[0][$f];
        }
        return 1;
    }
}

sub compute_reports_for_stream {
    my ( $self, $stream_name ) = @_;

    $self->init_reports_for_stream( $stream_name );
    my $sort_orders = $self->reports_by_sortorders( $stream_name );

    # Compute all the reports
    foreach my $rstruct ( @$sort_orders ) {
	$self->compute_reports( $stream_name, $rstruct );
    }
}

sub end_reports {
    my ( $self, $rstruct ) = @_;

    my $reports = $rstruct->{reports};
    my $filters = $rstruct->{filters};

    # Finalize the reports
    foreach my $report ( (@$reports,
			  map { @{$_->{reports}} } values %$filters ))
    {
	my $report_spec = $self->{report_spec_by_key}{$report};
	next unless defined $report_spec; # Skipped failed reports

	lr_info( "completing report '$report'" );
	eval {
	    $report_spec->calc_spec->end_report();
	};
	if ($@) {
	    lr_warn( "$@\n$report will be skipped");
	    $report_spec->mark_missing( "failed: $@" );
	    delete $self->{report_spec_by_key}{$report};
	}
    }
}

sub compute_reports {
    my ($self, $stream_name, $rstruct ) = @_;

    my $dlf_stream = eval {
        my $sort_spec = join( " ", @{$rstruct->{sort_fields}});
        $self->{dlf_store}->open_dlf_stream( $stream_name, "r", $sort_spec );
    };
    if ( $@ ) {
	lr_warn( $@ );
	$self->cancel_reports( $rstruct );
	return;
    }

    my $source_name;
    if ( @{$rstruct->{sort_fields}} ) {
	$source_name = "'" . $dlf_stream->name . "' sorted on fields: " .
	  $dlf_stream->sort_spec;
    } else {
	$source_name = "'" . $dlf_stream->name . "' unsorted";
    }

    lr_info( "computing reports from DLF source ", $source_name );

    my $reports = $rstruct->{reports};
    my $filters = $rstruct->{filters};
    my $record_count = $dlf_stream->nrecords;

    my $dlf;
    my $count = 0;
    while ( $dlf = $dlf_stream->read_dlf_aref ) {
	$count++;

	lr_info( sprintf( "%.2f%%", $count*100 / $record_count ),
		 " of DLF source $source_name processed" )
	  unless $count % 10_000;

	# Compute reports which aren't using a filter
	foreach my $report ( @$reports ) {
	    my $report_spec = $self->{report_spec_by_key}{$report};
	    next unless defined $report_spec; # Skipped failed reports
	    eval {
		$report_spec->calc_spec->update_report( $dlf );
	    };
	    if ( $@ ) {
		lr_warn( "$@\n$report will be skipped");
		$report_spec->mark_missing( "failed: $@" );
		delete $self->{report_spec_by_key}{$report};
	    }
	}

	# Compute each reports using a filter
	foreach my $f ( values %$filters) {
	    next unless $f->{filter}->( $dlf );

	    foreach my $report ( @{$f->{reports}} ) {
		my $report_spec = $self->{report_spec_by_key}{$report};
		next unless defined $report_spec; # Skipped failed reports
		eval {
		    $report_spec->calc_spec->update_report( $dlf );
		};
		if ( $@ ) {
		    lr_warn( "$@\n$report will be skipped");
		    $report_spec->mark_missing( "failed: $@" );
		    delete $self->{report_spec_by_key}{$report};
		}
	    }
	}
    }
    lr_info( "processed $count records in DLF source $source_name" );

    $self->end_reports( $rstruct );

    $dlf_stream->close;
}

# keep perl happy
1;

__END__

=pod

=head1 SEE ALSO

Lire::ReportConfig(3pm) lr_dlf2xml(1)

=head1 VERSION

$Id: ReportGenerator.pm,v 1.13 2003/04/18 21:48:04 flacoste Exp $

=head1 AUTHOR

Francis J. Lacoste <flacoste@logreport.org>

=head1 COPYRIGHT

Copyright (C) 2003 Stichting LogReport Foundation LogReport@LogReport.org

This file is part of Lire.

Lire is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program (see COPYING); if not, check with
http://www.gnu.org/copyleft/gpl.html or write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.

=cut

