package tests::ConfigSpecTest;

use strict;

use base qw/Test::Unit::TestSuite/;

use Lire::Config::Spec;
use Carp;

sub name {
    return "Configuration Specification Tests";
}

sub include_tests {
    return qw/tests::ConfigSpecCompoundTest tests::ConfigSpecListTest
              tests::ConfigSpecFileTest tests::ConfigSpecExecutableTest
              tests::ConfigSpecDirectoryTest tests::ConfigSpecBooleanTest
              tests::ConfigSpecStringTest tests::ConfigSpecIntegerTest
              tests::ConfigSpecSelectTest tests::ConfigSpecOptionTest
              tests::ConfigSpecConfigTest tests::ConfigSpecCommandTest
            /;
}

package tests::ConfigSpecTypeTest;

use base qw/Lire::Test::TestCase/;

sub set_up {
    my $self = $_[0];

    $self->{spec} = $self->type->new( name => "test_spec" );
}

sub tear_down {
    my $self = $_[0];

    # Since we aren't calling our parent's set_up, we shouldn't
    # call our parent's tear_down().
}

sub test_new {
    my $self = $_[0];

    my $type = $self->type;
    my $spec = $type->new( name => "parameter" );
    $self->assert_not_null( $spec, "$type->new returned undef" );
    $self->assert( UNIVERSAL::isa( $spec, $type ),
                   "$type->new returned value of wrong type: $spec"
                 );
    $self->assert_equals( "parameter", $spec->name );

    $self->assert_died( sub { $type->new() },
                        qr/missing 'name' parameter/ );
    $self->assert_died( sub { $type->new( name => 'wrong%name' )},
                        qr/name should only contain/ );
}

sub test_instance {
    my $self = $_[0];

    my $value = $self->{spec}->instance;
    $self->assert_not_null( $value, $self->type . "->instance() returned undef" );
    $self->assert( UNIVERSAL::isa( $value, $self->value_type ),
                   $self->type . "->instance() returned value of wrong type: $value" );
}

sub test_is_instance {
    my $self = $_[0];

    my $value = $self->{spec}->instance;

    $self->assert( $self->{spec}->is_instance( $value ),
                   "is_instance() returned false",
                 );
    $self->assert_died( sub { $self->{spec}->is_instance },
                        qr/missing 'instance' parameter/ );
    $self->assert_died( sub { $self->{spec}->is_instance( $self->{spec}) },
                        qr/'instance' parameter should be a Lire::Config::Value/ );

    my $bad_value = $self->type->new( name=> "bad_name" )->instance;
    $self->assert( ! $self->{spec}->is_instance( $bad_value ),
                   "is_instance() with value with different name should return false"
                 );

    my $clone = $self->type->new( name => $self->{spec}->name )->instance;
    $self->assert( ! $self->{spec}->is_instance( $clone ),
                   "is_instance() with different spec should die" );
}


sub test_description {
    my $self = $_[0];

    $self->assert_null( $self->{spec}->description,
                        "uninitialized description should be empty" );

    my $desc = "<para>Description</para>";
    $self->{spec}->description( $desc );
    $self->assert_equals( $desc, $self->{spec}->description );

    $self->assert_died( sub {$self->{spec}->description( "Description" ) },
                        qr/'description' parameter should be a DocBook/ );
}

package tests::ConfigSpecScalarTest;

use base qw/tests::ConfigSpecTypeTest/;

sub test_normalize {
    my $self = $_[0];

    $self->assert_null( $self->{spec}->normalize( undef ),
                        "normalize( undef ) should return undef",
                      );

    my %normalize_test_data = $self->normalize_test_data;
    while ( my ( $value, $expected ) = each %normalize_test_data ) {
        my $normalized = $self->{spec}->normalize( $value );
        $self->assert_equals( $expected, $normalized,
                              "normalize( ", $value, " should return ", $expected, ", not ", $normalized );
    }
}

sub test_is_valid {
    my $self = $_[0];

    my $warning = '';
    local $SIG{__WARN__} = sub { $warning .= join "", @_ };
    $self->assert( ! $self->{spec}->is_valid( undef ),
                   "is_valid( undef ) returned true" );
    $self->assert( ! $warning, "is_valid( undef ) had warnings: $warning" );
    foreach my $v ( $self->valid_test_data ) {
        $self->assert( $self->{spec}->is_valid( $v ),
                       $self->type . "->is_valid should return true for $v" );
        $self->assert( ! $warning, "is_valid( '$v' ) had warnings: $warning" );
    }

    foreach my $v ( $self->invalid_test_data ) {
        $self->assert( !$self->{spec}->is_valid( $v ),
                       $self->type . "->is_valid should return false for $v" );
        $self->assert( ! $warning, "is_valid( '$v' ) had warnings: $warning" );
    }
}

package tests::ConfigSpecCompoundTest;

use base qw/tests::ConfigSpecTypeTest/;

sub test_instance {
    my $self = $_[0];

    $self->assert_died( sub { $self->{spec}->instance() },
                        qr/not implemented/ );
}

sub test_is_instance {
    my $self = $_[0];

    my $dict = new Lire::Config::Dictionary( spec => $self->{spec} );
    $self->assert( $self->{spec}->is_instance( $dict ),
                   "is_instance() should be true " );
}

sub test_add {
    my $self = $_[0];

    $self->assert_equals( 0, scalar $self->{spec}->components);
    my $string = new Lire::Config::Spec::String( name => "test" );
    $self->{spec}->add( $string );
    $self->assert( $self->{spec}->has_component( "test" ),
                   $self->type . "->has_component returned false" );
    $self->assert( ! $self->{spec}->has_component( "Test" ),
                   $self->type . "->has_component returned true" );

    $self->assert_died( sub { $self->{spec}->add( $string ) },
                        qr/already contains a component named 'test'/ );
    $self->assert_equals( $self->{spec}->get( 'test' ), $string );

    my $int = new Lire::Config::Spec::Integer( name => "int" );
    $self->{spec}->add( $int );
    $self->assert_equals( $self->{spec}->get( "int" ), $int );

    $self->assert_died( sub {$self->{spec}->add( undef ) },
                        qr/missing 'type' parameter/ );

    $self->assert_died( sub {$self->{spec}->add( new Lire::Config::Scalar( spec => $self->{spec} ) ) },
                        qr/'type' parameter should be a Lire::Config::Spec::Type/ );

    $self->assert_deep_equals( [sort ( $int, $string ) ],
                               [ sort $self->{spec}->components ]
                             );
}

sub test_get {
    my $self = $_[0];

    $self->assert_died( sub { $self->{spec}->get( undef ) },
                        qr/missing 'name' parameter/ );

    $self->assert_died( sub { $self->{spec}->get( "param" ) },
                        qr/no component named 'param'/ );
}

sub type {
    return "Lire::Config::Spec::Compound";
}

package tests::ConfigSpecFileTest;

use base qw/tests::ConfigSpecScalarTest/;

sub type {
    return 'Lire::Config::Spec::File';
}

sub value_type {
    return 'Lire::Config::Scalar';
}

sub normalize_test_data {
    return (
            "/etc/passwd" => "/etc/passwd",
            "~/tmp" => "$ENV{HOME}/tmp",
           );
}

sub valid_test_data {
    return ( "/etc/passwd", __FILE__ );
}

sub invalid_test_data {
    return ( "~", "/no/such/file", "/dev/null" );
}

package tests::ConfigSpecExecutableTest;

use base qw/tests::ConfigSpecScalarTest/;

sub set_up {
    my $self = $_[0];
    $self->SUPER::set_up();

    $self->{old_PATH} = $ENV{PATH};
    $ENV{PATH} = '/bin:/sbin';
}

sub tear_down {
    my $self = $_[0];
    $self->SUPER::tear_down();


    $ENV{PATH} = defined $self->{old_PATH} ? $self->{old_PATH} : '';
}

sub type {
    return 'Lire::Config::Spec::Executable';
}

sub value_type {
    return 'Lire::Config::Scalar';
}

sub normalize_test_data {
    return (
            "sh" => "/bin/sh",
            "/bin/sh" => "/bin/sh",
            "echo" => "/bin/echo",
            "bin/echo" => "bin/echo",
            "/./bin/echo" => "/bin/echo",
           );
}

sub valid_test_data {
    return ( "sh", "/bin/sh", "/bin//cat" );
}

sub invalid_test_data {
    return ( "/bin/nosuchfile", "/etc", "/dev/null", "/etc/passwd" );
}

package tests::ConfigSpecDirectoryTest;

use base qw/tests::ConfigSpecScalarTest/;

use File::Basename;

sub type {
    return 'Lire::Config::Spec::Directory';
}

sub value_type {
    return 'Lire::Config::Scalar';
}

sub normalize_test_data {
    return (
            "/etc/passwd" => "/etc/passwd",
            "/bin/./../" => "/bin/..",
            "~" => "$ENV{HOME}",
           );
}

sub valid_test_data {
    return ( ".", "/bin", dirname( __FILE__ ) . "/../../" );
}

sub invalid_test_data {
    return ( "/bin/nosuchfile", __FILE__, "/dev/null", "/etc/passwd" );
}

package tests::ConfigSpecListTest;

use base qw/tests::ConfigSpecCompoundTest/;

sub type {
    return 'Lire::Config::Spec::List';
}

sub value_type {
    return 'Lire::Config::List';
}

sub test_instance {
    shift->tests::ConfigSpecTypeTest::test_instance();
}

sub test_is_instance {
    shift->tests::ConfigSpecTypeTest::test_is_instance();
}

package tests::ConfigSpecConfigTest;

use base qw/tests::ConfigSpecTypeTest/;

sub type {
    return 'Lire::Config::Spec::Config';
}

sub value_type {
    return 'Lire::Config::Dictionary';
}

sub test_new {
    my $self = $_[0];

    my $spec = $self->type->new();
    $self->assert_not_null( $spec, "new() returned undef" );
    $self->assert( UNIVERSAL::isa( $spec, $self->type ),
                   "new() returned value of wrong type: $spec"
                 );
}

package tests::ConfigSpecStringTest;

use base qw/tests::ConfigSpecScalarTest/;

sub type {
    return 'Lire::Config::Spec::String';
}

sub value_type {
    return 'Lire::Config::Scalar';
}

sub normalize_test_data {
    return ( test => "test",
             "another string" => "another string",
           );
}

sub valid_test_data {
    return ( "test", "string", 10 );
}

sub invalid_test_data {
    return ();
}

package tests::ConfigSpecIntegerTest;

use base qw/tests::ConfigSpecScalarTest/;

sub type {
    return 'Lire::Config::Spec::Integer';
}

sub value_type {
    return 'Lire::Config::Scalar';
}

sub normalize_test_data {
    return (
            0 => 0,
            -1_100_200 => "-1100200",
            1 => 1,
           );
}

sub valid_test_data {
    return ( 0, 1, -1, 1_200_300, -1_100, "-10", "-1_000_000" );
}

sub invalid_test_data {
    return ( 0.1, "two" );
}

package tests::ConfigSpecBooleanTest;

use base qw/tests::ConfigSpecScalarTest/;

sub type {
    return 'Lire::Config::Spec::Boolean';
}

sub value_type {
    return 'Lire::Config::Scalar';
}

sub normalize_test_data {
    return (
            '' => '',
            0 => '',
            1 => 1,
            enabled => 1,
            disabled => '',
            on => 1,
            off => '',
            Yes => 1,
            yes => 1,
            no => '',
            NO => '',
            F => '',
            FALSE => '',
            TRUE => 1,
            t => 1,
           );
}

sub valid_test_data {
    return ('', qw/0 1 enabled DISABLED Off oN Yes no True false t f/);
}

sub invalid_test_data {
    return qw/y n boolean/;
}

package tests::ConfigSpecSelectTest;

use base qw/tests::ConfigSpecScalarTest/;

sub set_up {
    my $self = $_[0];

    $self->SUPER::set_up;

    $self->{spec}->add( new Lire::Config::Spec::Option( name => "option_1" ) );
    $self->{spec}->add( new Lire::Config::Spec::Option( name => "option_2" ) );
    $self->{spec}->add( new Lire::Config::Spec::Option( name => "option_3" ) );
}

sub type {
    return "Lire::Config::Spec::Select";
}

sub value_type {
    return "Lire::Config::Scalar";
}

sub normalize_test_data {
    return (
            option_1 => "option_1",
            option_2 => "option_2",
            option_3 => "option_3",
            option_4 => undef,
            OPTION_1 => "option_1",
            OpTiOn_2 => "option_2",
           );
}

sub valid_test_data {
    return qw/option_1 Option_2 OPTION_3 option_3 option_2/;
}

sub invalid_test_data {
    return qw/option_4/;
}

sub test_add {
    my $self = $_[0];

    my $select = new Lire::Config::Spec::Select( name => "select" );
    $self->assert_equals( 0, scalar $select->options );

    my $option_1 = new Lire::Config::Spec::Option( name => "option_1" );
    $select->add( $option_1 );
    $self->assert_deep_equals( [ $option_1 ], [ $select->options ] );

    my $option_2 = new Lire::Config::Spec::Option( name => "option_2" );
    $select->add( $option_2 );
    $self->assert_deep_equals( [ sort ($option_1, $option_2) ],
                               [ sort $select->options ] );

    $self->assert_died( sub { $select->add( undef ) },
                        qr/missing 'option' parameter/ );
    $self->assert_died( sub { $select->add( new Lire::Config::Spec::Option( name => "OPTION_1" ) ) },
                        qr/already contains an option named 'option_1'/ );

    $self->assert_died( sub { $select->add( new Lire::Config::Spec::List( name => "OPTION_1" ) ) },
                              qr/'option' parameter should be a Lire::Config::Spec::Option instance/ );
}

package tests::ConfigSpecOptionTest;

use base qw/tests::ConfigSpecTypeTest/;

sub type {
    return "Lire::Config::Spec::Option";
}

sub test_instance  {
    my $self = $_[0];

    $self->assert_died( sub { $self->{spec}->instance() },
                        qr/not implemented/ );
}

sub test_is_instance {
    my $self = $_[0];

    my $scalar = new Lire::Config::Scalar( spec => new Lire::Config::Spec::Scalar( name => "test" ) );
    $self->assert( !$self->{spec}->is_instance( $scalar ),
                   "is_instance() should return false" );
}

package tests::ConfigSpecCommandTest;

use base qw/tests::ConfigSpecScalarTest/;

use File::Basename;

sub set_up {
    my $self = $_[0];
    $self->SUPER::set_up();

    $self->{old_PATH} = $ENV{PATH};
    $ENV{PATH} = '/bin:/sbin';
}

sub tear_down {
    my $self = $_[0];
    $self->SUPER::tear_down();


    $ENV{PATH} = defined $self->{old_PATH} ? $self->{old_PATH} : '';
}

sub type {
    return 'Lire::Config::Spec::Command';
}

sub value_type {
    return 'Lire::Config::Scalar';
}

sub normalize_test_data {
    return (
            "//bin//sh" => "/bin/sh",
            "sh" => "/bin/sh",
           );
}

sub valid_test_data {
    return ( "sh", "/bin/sh", "ls", "cat" );
}

sub invalid_test_data {
    return ( "/bin/nosuchfile", __FILE__, "/dev/null" );
}


1;
