package tests::DlfStreamTest;

use strict;

use base qw/ Lire::Test::TestCase /;

use File::Temp  'tempdir';
use File::Spec;
use Lire::DlfStore;
use Lire::DlfStream;
use Symbol;


my $schema_v1 = <<EOF;
<lire:dlf-schema superservice="test" timestamp="time_start"
 xmlns:lire="http://www.logreport.org/LDSML/">
<lire:field name="time_start" type="timestamp"/>
<lire:field name="field_1" type="string"/>
<lire:field name="field_2" type="string"/>
<lire:field name="int_3" type="int"/>
</lire:dlf-schema>
EOF

my $time_start = 900_000;
my $time_end = 1_000_000;
my @v1_fields = qw/time_start field_1 field_2 int_3/;
my @v1_dlf =
(
 { time_start => $time_start + 100,
   field_1 => "Field one",
   field_2 => "This is field 2",
   int_3 => 4,
 },
 { time_start => $time_start,
   field_1 => "field_1",
   field_2 => undef,
   int_3 => undef,
 },
 { time_start => $time_start + 1000,
   field_1 => "Day two",
   field_2 => "Another test",
   int_3 => undef,
 },
 { time_start => $time_end,
   field_1 => undef,
   field_2 => undef,
   int_3 => 1,
 },
);

# Remove the field_1 and add field_4, swap time_start and field_2
my @v2_fields = qw/field_2 time_start int_3 field_4/;
my $schema_v2 = <<EOF;
<lire:dlf-schema superservice="test" timestamp="time_start"
 xmlns:lire="http://www.logreport.org/LDSML/">
<lire:field name="field_2" type="string"/>
<lire:field name="time_start" type="timestamp"/>
<lire:field name="int_3" type="int"/>
<lire:field name="field_4" type="string"/>
</lire:dlf-schema>
EOF

sub new {
    my $self = shift()->SUPER::new( @_ );

    $self->{tmpdir} = tempdir( __PACKAGE__ . "XXXXXX",
                               CLEANUP => 1, TMPDIR => 1) ;
    $self->{store} = Lire::DlfStore->open( "$self->{tmpdir}/store", 1 );

    $self->{v1_dlf} = \@v1_dlf;
    $self->{v1_dlf_aref} = [];
    $self->create_aref( \@v1_fields, \@v1_dlf, $self->{v1_dlf_aref} );

    # Create the DLF v2 fields
    $self->{v2_dlf} = [];
    my $i=0;
    foreach my $dlf ( @v1_dlf ) {
        my $v2_dlf = { map { $_ => undef } @v2_fields};
        foreach my $k ( keys %$dlf ) {
            $v2_dlf->{$k} = $dlf->{$k} if grep { $_ eq $k } @v2_fields;
        }
        push @{$self->{v2_dlf}}, $v2_dlf;
    }

    $self->{v2_dlf_aref} = [];
    $self->create_aref( \@v2_fields, $self->{v2_dlf}, $self->{v2_dlf_aref} );

    $self;
}

sub create_aref {
    my ( $self, $fields, $dlf, $dlf_aref ) = @_;

    foreach my $r ( @$dlf ) {
        my $n = [];
        for (my $i=0; $i<@$fields; $i++) {
            $n->[$i] = $r->{$fields->[$i]};
        }
        push @$dlf_aref, $n;
    }
}

sub set_up {
    my $self = shift->SUPER::set_up();

    # Some tests might have changed the schema.
    $self->write_schema( "test", $schema_v1 );
    $self->{cfg}{lr_schemas_path} = [ $self->{tmpdir} ];
}

sub tear_down {
    my $self = shift->SUPER::tear_down();

    # Remove the possibly created test stream
    my $path = $self->{store}->storage_file( "dlf", "test" );
    unlink "$path.dlf", "$path.meta", "$path.stats";
}

sub write_schema {
    my ( $self, $name, $schema ) = @_;

    my $fh = gensym;
    open $fh, "> $self->{tmpdir}/$name.xml"
      or die "can't create temporary schema: $!\n";
    print $fh $schema;
    close $fh;
}

sub test_new_stream {
    my $self = $_[0];

    # Try to open a non-existent stream
    my $s = eval { new Lire::DlfStream( $self->{store}, "test", "r" ) };
    $self->assert_not_null( $@, "opening a non-existent stream should fail" );

    # Create
    $s = new Lire::DlfStream( $self->{store}, "test", "w" );
    $self->assert_not_null( $s, "creating the stream failed" );
    $self->assert_equals( 0, $s->nrecords,
                          "nrecords() should returns 0 on new stream" );
    $self->assert_null( $s->start_time,
                        "start_time() should returns undef on new stream" );
    $self->assert_null( $s->end_time,
                        "start_time() should returns undef on new stream" );
    $s->close;

    my $path = $self->{store}->storage_file( "dlf", "test" );
    $self->assert( -f "$path.dlf", "no DLF file created" );
    $self->assert( -f "$path.meta", "no DLF meta file created" );
    $self->assert( -f "$path.stats", "no DLF stats file created" );

    # Open now in read mode
    $s = new Lire::DlfStream( $self->{store}, "test", "r" );
    $self->assert_not_null( $s, "failed to reopen new stream" );
    $self->assert_equals( 0, $s->nrecords );
}

sub test_read_write_dlf {
    my $self = $_[0];

    my $s = new Lire::DlfStream( $self->{store}, "test", "w" );
    $self->assert_not_null( $s, "creating the stream failed" );
    foreach my $dlf ( @{$self->{v1_dlf}} ) {
        $s->write_dlf( $dlf );
    }

    eval { $s->read_dlf };
    $self->assert_not_null( $@, "reading from 'w' stream should fail" );
    $s->close;

    # Reopen
    $s = new Lire::DlfStream( $self->{store}, "test", "r" );
    eval { $s->write_dlf( $self->{v1_dlf}[0] ) };
    $self->assert_not_null( $@, "writing to 'r' stream should fail" );

    my @dlf = ();
    while ( defined( $_ = $s->read_dlf ) ) {
        push @dlf, $_;
    }
    $self->assert_deep_equals( $self->{v1_dlf}, \@dlf );
    $s->close;

}

sub test_read_dlf_aref {
    my $self = $_[0];

    my $s = new Lire::DlfStream( $self->{store}, "test", "w" );
    $self->assert_not_null( $s, "creating the stream failed" );
    foreach my $dlf ( @{$self->{v1_dlf}} ) {
        $s->write_dlf( $dlf );
    }

    # Reopen
    $s = new Lire::DlfStream( $self->{store}, "test", "r" );
    eval { $s->write_dlf( $self->{v1_dlf}[0] ) };
    $self->assert_not_null( $@, "writing to 'r' stream should fail" );

    my @dlf = ();
    while ( defined( $_ = $s->read_dlf_aref ) ) {
        push @dlf, $_;
    }
    $self->assert_deep_equals( $self->{v1_dlf_aref}, \@dlf );
    $s->close;
}

sub test_dlf_stats {
    my $self = $_[0];

    my $s = new Lire::DlfStream( $self->{store}, "test", "w" );
    $self->assert_not_null( $s, "creating the stream failed" );
    my $i=0;
    foreach my $dlf ( @{$self->{v1_dlf}} ) {
        $s->write_dlf( $dlf );
        $i++;
        $self->assert_equals( $i, $s->nrecords );
    }
    $self->assert_equals( $time_start, $s->start_time,
                          "start_time() differs" );
    $self->assert_equals( $time_end, $s->end_time,
                          "end_time() differs" );
    $s->close;

    my $stats_file = $self->{store}->storage_file( "dlf", "test.stats" );
    $self->assert( -f $stats_file, "cannot find DLF stats file" );

    # Reopen and check the stats
    $s = new Lire::DlfStream( $self->{store}, "test", "r" );
    $self->assert_not_null( $s, "reopening the stream failed" );

    $self->assert_equals( $i, $s->nrecords );
    $self->assert_equals( $time_start, $s->start_time );
    $self->assert_equals( $time_end, $s->end_time );

    $s->close;

    # Test rebuild of the stats
    unlink $stats_file;

    $s = new Lire::DlfStream( $self->{store}, "test", "r" );
    $self->assert_not_null( $s, "reopening the stream failed" );

    $self->assert_equals( $i, $s->nrecords );
    $self->assert_equals( $time_start, $s->start_time );
    $self->assert_equals( $time_end, $s->end_time );
    $s->close;

    $self->assert( -f $stats_file, "failed to recreate the stats file" );
}

sub test_from_to_dlf {
    my $self = $_[0];

    my $fields = [ "time_start", "field1", "field2", "int3" ];
    my $dlf = { time_start => 90000,
                field1 => undef,
                field2 => "string 2",
                int3 => 0
              };
    my $line = "90000\tLIRE_NA\tstring 2\t0";
    my $s = Lire::DlfStream::_to_dlf( $fields, $dlf );
    $self->assert_equals( $line, $s );

    my $d = Lire::DlfStream::_from_dlf( $fields, $line );
    $self->assert_deep_equals( $dlf, $d );
}

sub create_test_stream {
    my $self = $_[0];

    my $s = new Lire::DlfStream( $self->{store}, "test", "w" );
    $self->assert_not_null( $s, "creating the stream failed" );
    foreach my $dlf ( @{$self->{v1_dlf}} ) {
        $s->write_dlf( $dlf );
    }
    $s->close;
}

sub migrate_schema_v2 {
    my $self = $_[0];

    $self->write_schema( "test", $schema_v2 );
    # Make sure that the changes are noticed
    my $now = time + 5;
    utime $now, $now, "$self->{tmpdir}/test.xml";
}

sub test_read_dlf_migrate {
    my $self = $_[0];

    $self->create_test_stream;

    $self->migrate_schema_v2;

    my $s = new Lire::DlfStream( $self->{store}, "test", "r" );

    my @dlf = ();
    while (defined($_ = $s->read_dlf)) {
        push @dlf, $_;
    }
    $self->assert_deep_equals( $self->{v2_dlf}, \@dlf);
}

sub test_read_dlf_aref_migrate {
    my $self = $_[0];

    $self->create_test_stream;
    $self->migrate_schema_v2;

    my $s = new Lire::DlfStream( $self->{store}, "test", "r" );

    my @dlf = ();
    while (defined($_ = $s->read_dlf_aref)) {
        push @dlf, $_;
    }
    $self->assert_deep_equals( $self->{v2_dlf_aref}, \@dlf);
}

sub test_write_dlf_migrate {
    my $self = $_[0];

    $self->create_test_stream;
    $self->migrate_schema_v2;

    # Migration should be done transparently
    my $s = new Lire::DlfStream( $self->{store}, "test", "w" );
    # Write another one just to make sure
    my @v2_cmp = ( @{$self->{v2_dlf}}, $self->{v2_dlf}[0] );
    $s->write_dlf( $self->{v2_dlf}[0] );
    $self->assert_equals( @{$self->{v2_dlf}} + 1, $s->nrecords );
    $s->close;

    $s = new Lire::DlfStream( $self->{store}, "test", "r" );
    my @dlf = ();
    while (defined($_ = $s->read_dlf)) {
        push @dlf, $_;
    }

    $self->assert_deep_equals( \@v2_cmp, \@dlf);
}

sub test_sorted_time_start {
    my $self = $_[0];

    $self->create_test_stream;

    # Sort on time_start
    my $s = new Lire::DlfStream( $self->{store}, "test", "r", "time_start" );
    $self->assert_not_null( $s, "new() returned undef" );
    my @dlf = ();
    while (defined($_ = $s->read_dlf)) {
        push @dlf, $_;
    }
    $s->close;
}

sub test_rsorted_time_start {
    my $self = $_[0];

    $self->create_test_stream;

    # Sort on time_start
    my $s = new Lire::DlfStream( $self->{store}, "test", "r", "-time_start" );
    $self->assert_not_null( $s, "new() returned undef" );
    my @dlf = ();
    while (defined($_ = $s->read_dlf)) {
        push @dlf, $_;
    }
    $s->close;

    my @sorted_dlf = sort { $b->{time_start} <=> $a->{time_start} } @{$self->{v1_dlf}};
    $self->assert_deep_equals( \@sorted_dlf, \@dlf);
}

sub test_sorted_mkeys {
    my $self = $_[0];

    $self->create_test_stream;

    # Sort on time_start
    my $s = new Lire::DlfStream( $self->{store}, "test", "r", "field_1 -time_start" );
    $self->assert_not_null( $s, "new() returned undef" );
    my @dlf = ();
    while (defined($_ = $s->read_dlf)) {
        push @dlf, $_;
    }
    $s->close;

    # Some field_1 are undef ->
    no warnings 'uninitialized';

    my @sorted_dlf = sort { $a->{field_1} cmp $b->{field_1} ||
                            $b->{time_start} <=> $a->{time_start} } @{$self->{v1_dlf}};
    $self->assert_deep_equals( \@sorted_dlf, \@dlf);
}

sub test_sorted_migrate {
    my $self = $_[0];

    $self->create_test_stream;
    $self->migrate_schema_v2;

    # Sort on time_start
    my $s = new Lire::DlfStream( $self->{store}, "test", "r", "time_start" );
    $self->assert_not_null( $s, "new() returned undef" );
    my @dlf = ();
    while (defined($_ = $s->read_dlf)) {
        push @dlf, $_;
    }
    $s->close;

    my @sorted_dlf = sort { $a->{time_start} <=> $b->{time_start} } @{$self->{v2_dlf}};
    $self->assert_deep_equals( \@sorted_dlf, \@dlf);
}


1;
